from __future__ import annotations

import functools
import importlib
import os
import re
import stat
import typing as t

import click
import jsonschema
from click._compat import open_stream

C = t.TypeVar("C", bound=t.Callable[..., t.Any])


def _shim_click_8_2_get_metavar(func: C) -> C:
    @functools.wraps(func)
    def wrapper(*args: t.Any, **kwargs: t.Any) -> None:
        if len(args) > 1 or "ctx" in kwargs:
            return func(*args, **kwargs)
        return func(*args, ctx=None, **kwargs)

    return wrapper  # type: ignore[return-value]


class CommaDelimitedList(click.ParamType):
    name = "comma_delimited"

    def __init__(
        self,
        *,
        convert_values: t.Callable[[str], str] | None = None,
        choices: t.Iterable[str] | None = None,
    ) -> None:
        super().__init__()
        self.convert_values = convert_values
        self.choices = list(choices) if choices is not None else None

    @_shim_click_8_2_get_metavar
    def get_metavar(self, param: click.Parameter, ctx: click.Context | None) -> str:
        if self.choices is not None:
            return "{" + ",".join(self.choices) + "}"
        return "TEXT,TEXT,..."

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> list[str]:
        value = super().convert(value, param, ctx)

        # if `--foo` is a comma delimited list and someone passes
        # `--foo ""`, take that as `foo=[]` rather than foo=[""]
        resolved = value.split(",") if value else []

        if self.convert_values is not None:
            resolved = [self.convert_values(x) for x in resolved]

        if self.choices is not None:
            bad_values = [x for x in resolved if x not in self.choices]
            if bad_values:
                self.fail(
                    f"the values {bad_values} were not valid choices",
                    param=param,
                    ctx=ctx,
                )

        return resolved


class ValidatorClassName(click.ParamType):
    name = "validator"

    def convert(
        self, value: str, param: click.Parameter | None, ctx: click.Context | None
    ) -> type[jsonschema.protocols.Validator]:
        """
        Use a colon-based parse to split this up and do the import with importlib.
        This method is inspired by pkgutil.resolve_name and uses the newer syntax
        documented there.

        pkgutil supports both
            W(.W)*
        and
            W(.W)*:(W(.W)*)?
        as patterns, but notes that the first one is for backwards compatibility only.
        The second form is preferred because it clarifies the division between the
        importable name and any namespaced path to an object or class.

        As a result, only one import is needed, rather than iterative imports over the
        list of names.
        """
        value = super().convert(value, param, ctx)
        pattern = re.compile(
            r"^(?P<pkg>(?!\d)(\w+)(\.(?!\d)(\w+))*):"
            r"(?P<cls>(?!\d)(\w+)(\.(?!\d)(\w+))*)$"
        )
        m = pattern.match(value)
        if m is None:
            self.fail(
                f"'{value}' is not a valid specifier in '<package>:<class>' form",
                param,
                ctx,
            )
        pkg = m.group("pkg")
        classname = m.group("cls")
        try:
            result: t.Any = importlib.import_module(pkg)
        except ImportError as e:
            self.fail(f"'{pkg}' was not an importable module. {str(e)}", param, ctx)
        try:
            for part in classname.split("."):
                result = getattr(result, part)
        except AttributeError as e:
            self.fail(
                f"'{classname}' was not resolvable to a class in '{pkg}'. {str(e)}",
                param,
                ctx,
            )

        if not isinstance(result, type):
            self.fail(f"'{classname}' in '{pkg}' is not a class", param, ctx)

        return t.cast(t.Type[jsonschema.protocols.Validator], result)


class CustomLazyFile(click.utils.LazyFile):
    def __init__(
        self,
        filename: str | os.PathLike[str],
        mode: str = "r",
        encoding: str | None = None,
        errors: str | None = "strict",
        atomic: bool = False,
    ) -> None:
        self.name: str = os.fspath(filename)
        self.mode = mode
        self.encoding = encoding
        self.errors = errors
        self.atomic = atomic
        self._f: t.IO[t.Any] | None
        self.should_close: bool

        if self.name == "-":
            self._f, self.should_close = open_stream(filename, mode, encoding, errors)
        else:
            if "r" in mode and not stat.S_ISFIFO(os.stat(filename).st_mode):
                # Open and close the file in case we're opening it for
                # reading so that we can catch at least some errors in
                # some cases early.
                open(filename, mode).close()
            self._f = None
            self.should_close = True


class LazyBinaryReadFile(click.File):
    def convert(
        self,
        value: str | os.PathLike[str] | t.IO[t.Any],
        param: click.Parameter | None,
        ctx: click.Context | None,
    ) -> t.IO[bytes]:
        if hasattr(value, "read") or hasattr(value, "write"):
            return t.cast(t.IO[bytes], value)

        value_: str | os.PathLike[str] = t.cast("str | os.PathLike[str]", value)

        lf = CustomLazyFile(value_, mode="rb")
        if ctx is not None:
            ctx.call_on_close(lf.close_intelligently)
        return t.cast(t.IO[bytes], lf)



from __future__ import annotations

import os
import textwrap
import typing as t

import click
import jsonschema

from ..catalog import CUSTOM_SCHEMA_NAMES, SCHEMA_CATALOG
from ..checker import SchemaChecker
from ..formats import KNOWN_FORMATS
from ..instance_loader import InstanceLoader
from ..parsers import SUPPORTED_FILE_FORMATS
from ..regex_variants import RegexImplementation, RegexVariantName
from ..reporter import REPORTER_BY_NAME, Reporter
from ..schema_loader import (
    BuiltinSchemaLoader,
    MetaSchemaLoader,
    SchemaLoader,
    SchemaLoaderBase,
)
from ..transforms import TRANSFORM_LIBRARY
from .param_types import CommaDelimitedList, LazyBinaryReadFile, ValidatorClassName
from .parse_result import ParseResult, SchemaLoadingMode

BUILTIN_SCHEMA_NAMES = [f"vendor.{k}" for k in SCHEMA_CATALOG.keys()] + [
    f"custom.{k}" for k in CUSTOM_SCHEMA_NAMES
]
BUILTIN_SCHEMA_CHOICES = (
    BUILTIN_SCHEMA_NAMES + list(SCHEMA_CATALOG.keys()) + CUSTOM_SCHEMA_NAMES
)


def set_color_mode(ctx: click.Context, param: str, value: str) -> None:
    if "NO_COLOR" in os.environ:
        ctx.color = False
    else:
        ctx.color = {
            "auto": None,
            "always": True,
            "never": False,
        }[value]


def pretty_helptext_list(values: list[str] | tuple[str, ...]) -> str:
    return textwrap.indent(
        "\n".join(
            textwrap.wrap(
                ", ".join(values),
                width=75,
                break_long_words=False,
                break_on_hyphens=False,
            ),
        ),
        "    ",
    )


@click.command(
    "check-jsonschema",
    help="""\
Check JSON and YAML files against a JSON Schema.

The schema is specified either with '--schemafile' or with '--builtin-schema'.

'check-jsonschema' supports format checks with appropriate libraries installed,
including the following formats by default:
    date, date-time, email, ipv4, ipv6, regex, uuid

\b
For handling of regexes, there are multiple modes which can be specified with
'--regex-variant':
    default    |  use ECMAScript regex syntax (via regress)
    nonunicode |  use ECMAScript regex syntax, but in non-unicode mode (via regress)
    python     |  use python regex syntax

\b
The '--builtin-schema' flag supports the following schema names:
"""
    + pretty_helptext_list(BUILTIN_SCHEMA_NAMES)
    + """\

\b
The '--disable-formats' flag supports the following formats:
"""
    + pretty_helptext_list(KNOWN_FORMATS),
)
@click.help_option("-h", "--help")
@click.version_option()
@click.option(
    "--schemafile",
    help=(
        "The path to a file containing the JSON Schema to use or an "
        "HTTP(S) URI for the schema. If a remote file is used, "
        "it will be downloaded and cached locally based on mtime. "
        "Use '-' for stdin."
    ),
    metavar="[PATH|URI]",
)
@click.option(
    "--base-uri",
    help=(
        "Override the base URI for the schema. The default behavior is to "
        "follow the behavior specified by the JSON Schema spec, which is to "
        "prefer an explicit '$id' and failover to the retrieval URI."
    ),
)
@click.option(
    "--builtin-schema",
    help="The name of an internal schema to use for '--schemafile'",
    type=click.Choice(BUILTIN_SCHEMA_CHOICES, case_sensitive=False),
    metavar="BUILTIN_SCHEMA_NAME",
)
@click.option(
    "--check-metaschema",
    is_flag=True,
    help=(
        "Instead of validating the instances against a schema, treat each file as a "
        "schema and validate them under their matching metaschemas."
    ),
)
@click.option(
    "--no-cache",
    is_flag=True,
    help="Disable schema caching. Always download remote schemas.",
)
@click.option(
    "--cache-filename", help="Deprecated. This option no longer has any effect."
)
@click.option(
    "--disable-formats",
    multiple=True,
    help=(
        "Disable specific format checks in the schema. "
        "Pass '*' to disable all format checks."
    ),
    type=CommaDelimitedList(choices=("*", *KNOWN_FORMATS)),
    metavar="{*|FORMAT,FORMAT,...}",
)
@click.option(
    "--format-regex",
    hidden=True,
    help="Legacy name for `--regex-variant`.",
    default=None,
    type=click.Choice([x.value for x in RegexVariantName], case_sensitive=False),
)
@click.option(
    "--regex-variant",
    help=(
        "Name of which regex dialect should be used for format checking "
        "and 'pattern' matching."
    ),
    default=None,
    type=click.Choice([x.value for x in RegexVariantName], case_sensitive=False),
)
@click.option(
    "--default-filetype",
    help="A default filetype to assume when a file's type is not detected",
    default="json",
    show_default=True,
    type=click.Choice(SUPPORTED_FILE_FORMATS, case_sensitive=True),
)
@click.option(
    "--force-filetype",
    help="Force a file type to use when parsing instance files",
    type=click.Choice(SUPPORTED_FILE_FORMATS, case_sensitive=True),
)
@click.option(
    "--traceback-mode",
    help=(
        "Set the mode of presentation for error traces. "
        "Defaults to shortened tracebacks."
    ),
    type=click.Choice(("full", "short")),
    default="short",
)
@click.option(
    "--data-transform",
    help=(
        "Select a builtin transform which should be applied to instancefiles before "
        "they are checked."
    ),
    type=click.Choice(tuple(TRANSFORM_LIBRARY.keys())),
)
@click.option(
    "--fill-defaults",
    help=(
        "Autofill 'default' values prior to validation. "
        "This may conflict with certain third-party validators used with "
        "'--validator-class'"
    ),
    is_flag=True,
)
@click.option(
    "--validator-class",
    help=(
        "The fully qualified name of a python validator to use in place of "
        "the 'jsonschema' library validators, in the form of '<package>:<class>'. "
        "The validator must be importable in the same environment where "
        "'check-jsonschema' is run."
    ),
    type=ValidatorClassName(),
)
@click.option(
    "-o",
    "--output-format",
    help="Which output format to use.",
    type=click.Choice(tuple(REPORTER_BY_NAME.keys()), case_sensitive=False),
    default="text",
)
@click.option(
    "--color",
    help="Force or disable colorized output. Defaults to autodetection.",
    default="auto",
    type=click.Choice(("auto", "always", "never")),
    callback=set_color_mode,
    expose_value=False,
)
@click.option(
    "-v",
    "--verbose",
    help=(
        "Increase output verbosity. On validation errors, this may be especially "
        "useful when oneOf or anyOf is used in the schema."
    ),
    count=True,
)
@click.option(
    "-q",
    "--quiet",
    help="Reduce output verbosity",
    count=True,
)
@click.argument(
    "instancefiles", required=True, nargs=-1, type=LazyBinaryReadFile("rb", lazy=True)
)
def main(
    *,
    schemafile: str | None,
    builtin_schema: str | None,
    base_uri: str | None,
    check_metaschema: bool,
    no_cache: bool,
    cache_filename: str | None,
    disable_formats: tuple[list[str], ...],
    format_regex: t.Literal["python", "nonunicode", "default"] | None,
    regex_variant: t.Literal["python", "nonunicode", "default"] | None,
    default_filetype: t.Literal["json", "yaml", "toml", "json5"],
    force_filetype: t.Literal["json", "yaml", "toml", "json5"] | None,
    traceback_mode: t.Literal["full", "short"],
    data_transform: t.Literal["azure-pipelines", "gitlab-ci"] | None,
    fill_defaults: bool,
    validator_class: type[jsonschema.protocols.Validator] | None,
    output_format: t.Literal["text", "json"],
    verbose: int,
    quiet: int,
    instancefiles: tuple[t.IO[bytes], ...],
) -> None:
    args = ParseResult()

    args.set_regex_variant(regex_variant, legacy_opt=format_regex)

    args.set_schema(schemafile, builtin_schema, check_metaschema)
    args.set_validator(validator_class)

    args.base_uri = base_uri
    args.instancefiles = instancefiles

    normalized_disable_formats: tuple[str, ...] = tuple(
        f for sublist in disable_formats for f in sublist
    )
    if "*" in normalized_disable_formats:
        args.disable_all_formats = True
    else:
        args.disable_formats = normalized_disable_formats

    args.disable_cache = no_cache
    args.default_filetype = default_filetype
    args.force_filetype = force_filetype
    args.fill_defaults = fill_defaults
    if data_transform is not None:
        args.data_transform = TRANSFORM_LIBRARY[data_transform]

    # verbosity behavior:
    # - default is 1
    # - count '-v'
    # - subtract count of '-q'
    args.verbosity = 1 + verbose - quiet
    args.traceback_mode = traceback_mode
    args.output_format = output_format

    execute(args)


# separate parsing from execution for simpler mocking for unit tests


def build_schema_loader(args: ParseResult) -> SchemaLoaderBase:
    if args.schema_mode == SchemaLoadingMode.metaschema:
        return MetaSchemaLoader(base_uri=args.base_uri)
    elif args.schema_mode == SchemaLoadingMode.builtin:
        assert args.schema_path is not None
        return BuiltinSchemaLoader(args.schema_path, base_uri=args.base_uri)
    elif args.schema_mode == SchemaLoadingMode.filepath:
        assert args.schema_path is not None
        return SchemaLoader(
            args.schema_path,
            disable_cache=args.disable_cache,
            base_uri=args.base_uri,
            validator_class=args.validator_class,
        )
    else:
        raise NotImplementedError("no valid schema option provided")


def build_instance_loader(args: ParseResult) -> InstanceLoader:
    return InstanceLoader(
        args.instancefiles,
        default_filetype=args.default_filetype,
        force_filetype=args.force_filetype,
        data_transform=args.data_transform,
    )


def build_reporter(args: ParseResult) -> Reporter:
    cls = REPORTER_BY_NAME[args.output_format]
    return cls(verbosity=args.verbosity)


def build_checker(args: ParseResult) -> SchemaChecker:
    schema_loader = build_schema_loader(args)
    instance_loader = build_instance_loader(args)
    reporter = build_reporter(args)
    return SchemaChecker(
        schema_loader,
        instance_loader,
        reporter,
        format_opts=args.format_opts,
        regex_impl=RegexImplementation(args.regex_variant),
        traceback_mode=args.traceback_mode,
        fill_defaults=args.fill_defaults,
    )


def execute(args: ParseResult) -> None:
    checker = build_checker(args)
    ret = checker.run()
    click.get_current_context().exit(ret)



from __future__ import annotations

import datetime
import sys
import typing as t

if sys.version_info < (3, 11):
    import tomli as toml_implementation
else:
    import tomllib as toml_implementation


def _normalize(data: t.Any) -> t.Any:
    """
    Normalize TOML data to fit the requirements to be JSON-encodeable.

    Currently this applies the following transformations:

        offset-aware datetime.datetime values are converted to strings using isoformat()
        naive datetime.datetime values are converted to strings using isoformat() + "Z"

        offset-aware datetime.time values are converted to strings using isoformat()
        naive datetime.time values are converted to strings using isoformat() + "Z"

        datetime.date values are converted to strings using isoformat()
    """
    if isinstance(data, dict):
        return {k: _normalize(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_normalize(x) for x in data]
    else:
        # python's datetime will format to an ISO partial time when handling a naive
        # time/datetime , but JSON Schema format validation specifies that date-time is
        # taken from RFC3339, which defines "date-time" as including 'Z|offset'
        # the specification for "time" is less clear because JSON Schema does not specify
        # which RFC3339 definition should be used, and the RFC has no format named "time",
        # only "full-time" (with Z|offset) and "partial-time" (no offset)
        #
        # rfc3339_validator (used by 'jsonschema') requires the offset, so we will do the
        # same
        if isinstance(data, datetime.datetime) or isinstance(data, datetime.time):
            if data.tzinfo is None:
                return data.isoformat() + "Z"
            return data.isoformat()
        elif isinstance(data, datetime.date):
            return data.isoformat()
        return data


ParseError: type[Exception] = toml_implementation.TOMLDecodeError


def load(stream: t.IO[bytes]) -> t.Any:
    data = toml_implementation.load(stream)
    return _normalize(data)



import json

import pytest

from check_jsonschema.parsers.json5 import ENABLED as JSON5_ENABLED

SIMPLE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {
        "title": {"type": "string"},
    },
    "additionalProperties": False,
}
PASSING_DOCUMENT = """
// a comment
{"title": "doc one"}
"""
FAILING_DOCUMENT = """
// a comment
{"title": 2}
"""


@pytest.mark.skipif(not JSON5_ENABLED, reason="test requires json5")
@pytest.mark.parametrize("passing_data", [True, False])
def test_json5_filetype_forced_on_json_suffixed_instance(
    run_line, tmp_path, passing_data
):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(SIMPLE_SCHEMA))

    doc = tmp_path / "doc.json"
    if passing_data:
        doc.write_text(PASSING_DOCUMENT)
    else:
        doc.write_text(FAILING_DOCUMENT)

    result = run_line(
        [
            "check-jsonschema",
            "--force-filetype",
            "json5",
            "--schemafile",
            str(schemafile),
            str(doc),
        ]
    )
    assert result.exit_code == (0 if passing_data else 1)

    # but even in the passing case, a rerun without the force flag will fail
    if passing_data:
        result_without_filetype = run_line(
            [
                "check-jsonschema",
                "--schemafile",
                str(schemafile),
                str(doc),
            ]
        )
        assert result_without_filetype.exit_code == 1



from __future__ import annotations

import typing as t
import warnings

import ruamel.yaml

ParseError = ruamel.yaml.YAMLError


def construct_yaml_implementation(
    typ: str = "safe", pure: bool = False
) -> ruamel.yaml.YAML:
    implementation = ruamel.yaml.YAML(typ=typ, pure=pure)

    # workaround global state
    # see: https://sourceforge.net/p/ruamel-yaml/tickets/341/
    class GeneratedSafeConstructor(ruamel.yaml.SafeConstructor):
        pass

    implementation.Constructor = GeneratedSafeConstructor

    # ruamel.yaml parses timestamp values into datetime.datetime values
    # however, JSON does not support native datetimes, so JSON Schema formats for
    # dates apply to strings
    # Turn off this feature, instructing the parser to load datetimes as strings
    implementation.constructor.yaml_constructors["tag:yaml.org,2002:timestamp"] = (
        implementation.constructor.yaml_constructors["tag:yaml.org,2002:str"]
    )

    return implementation


def _normalize(data: t.Any) -> t.Any:
    """
    Normalize YAML data to fit the requirements to be JSON-encodeable.

    Currently this applies the following transformation:
        dict keys are converted to strings

    Additional tweaks can be added in this layer in the future if necessary.
    """
    if isinstance(data, dict):
        return {str(k): _normalize(v) for k, v in data.items()}
    elif isinstance(data, list):
        return [_normalize(x) for x in data]
    else:
        return data


_data_sentinel = object()


def impl2loader(
    primary: ruamel.yaml.YAML, *fallbacks: ruamel.yaml.YAML
) -> t.Callable[[t.IO[bytes]], t.Any]:
    def load(stream: t.IO[bytes]) -> t.Any:
        stream_bytes = stream.read()
        lasterr: ruamel.yaml.YAMLError | None = None
        data: t.Any = _data_sentinel
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ruamel.yaml.error.ReusedAnchorWarning)
            for impl in [primary] + list(fallbacks):
                try:
                    data = impl.load(stream_bytes)
                except ruamel.yaml.YAMLError as e:
                    lasterr = e
                else:
                    break
        if data is _data_sentinel and lasterr is not None:
            raise lasterr
        return _normalize(data)

    return load



import os
import platform
import sys

import pytest

from check_jsonschema.utils import filename2path


@pytest.mark.skipif(
    not (platform.system() == "Linux"), reason="test requires /proc/self/ mechanism"
)
@pytest.mark.skipif(sys.version_info < (3, 8), reason="test uses os.memfd_create")
@pytest.mark.parametrize("use_pid_in_path", (True, False))
def test_filename2path_on_memfd(use_pid_in_path):
    """
    create a memory file descriptor with a path in /proc/self/fd/
    and then attempt to resolve that to an absolute Path object
    the end result should be untouched

    pathlib behavior is, for example,

    >>> pathlib.Path("/proc/self/fd/4").resolve()
    PosixPath('/memfd:myfd (deleted)')
    """
    testfd = os.memfd_create("test_filename2path")
    try:
        pid = os.getpid() if use_pid_in_path else "self"
        filename = f"/proc/{pid}/fd/{testfd}"
        path = filename2path(filename)

        assert str(path) == filename
    finally:
        os.close(testfd)



from __future__ import annotations

import calendar
import contextlib
import hashlib
import io
import os
import platform
import shutil
import tempfile
import time
import typing as t

import requests

_LASTMOD_FMT = "%a, %d %b %Y %H:%M:%S %Z"


def _base_cache_dir() -> str | None:
    sysname = platform.system()

    # on windows, try to get the appdata env var
    # this *could* result in cache_dir=None, which is fine, just skip caching in
    # that case
    if sysname == "Windows":
        cache_dir = os.getenv("LOCALAPPDATA", os.getenv("APPDATA"))
    # macOS -> app support dir
    elif sysname == "Darwin":
        cache_dir = os.path.expanduser("~/Library/Caches")
    # default for unknown platforms, namely linux behavior
    # use XDG env var and default to ~/.cache/
    else:
        cache_dir = os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache"))

    return cache_dir


def _resolve_cache_dir(dirname: str) -> str | None:
    cache_dir = _base_cache_dir()
    if cache_dir:
        cache_dir = os.path.join(cache_dir, "check_jsonschema", dirname)
    return cache_dir


def _lastmod_from_response(response: requests.Response) -> float:
    try:
        return calendar.timegm(
            time.strptime(response.headers["last-modified"], _LASTMOD_FMT)
        )
    # OverflowError: time outside of platform-specific bounds
    # ValueError: malformed/unparseable
    # LookupError: no such header
    except (OverflowError, ValueError, LookupError):
        return 0.0


def _get_request(
    file_url: str, *, response_ok: t.Callable[[requests.Response], bool]
) -> requests.Response:
    num_retries = 2
    r: requests.Response | None = None
    for _attempt in range(num_retries + 1):
        try:
            r = requests.get(file_url, stream=True)
        except requests.RequestException as e:
            if _attempt == num_retries:
                raise FailedDownloadError("encountered error during download") from e
            continue
        if r.ok and response_ok(r):
            return r
    assert r is not None
    raise FailedDownloadError(
        f"got response with status={r.status_code}, retries exhausted"
    )


def _atomic_write(dest: str, content: bytes) -> None:
    # download to a temp file and then move to the dest
    # this makes the download safe if run in parallel (parallel runs
    # won't create a new empty file for writing and cause failures)
    fp = tempfile.NamedTemporaryFile(mode="wb", delete=False)
    fp.write(content)
    fp.close()
    shutil.copy(fp.name, dest)
    os.remove(fp.name)


def _cache_hit(cachefile: str, response: requests.Response) -> bool:
    # no file? miss
    if not os.path.exists(cachefile):
        return False

    # compare mtime on any cached file against the remote last-modified time
    # it is considered a hit if the local file is at least as new as the remote file
    local_mtime = os.path.getmtime(cachefile)
    remote_mtime = _lastmod_from_response(response)
    return local_mtime >= remote_mtime


def url_to_cache_filename(ref_url: str) -> str:
    """
    Given a schema URL, convert it to a filename for caching in a cache dir.

    Rules are as follows:
    - the base filename is an sha256 hash of the URL
    - if the filename ends in an extension (.json, .yaml, etc) that extension
      is appended to the hash

    Preserving file extensions preserves the extension-based logic used for parsing, and
    it also helps a local editor (browsing the cache) identify filetypes.
    """
    filename = hashlib.sha256(ref_url.encode()).hexdigest()
    if "." in (last_part := ref_url.rpartition("/")[-1]):
        _, _, extension = last_part.rpartition(".")
        filename = f"{filename}.{extension}"
    return filename


class FailedDownloadError(Exception):
    pass


class CacheDownloader:
    def __init__(self, cache_dir: str, *, disable_cache: bool = False) -> None:
        self._cache_dir = _resolve_cache_dir(cache_dir)
        self._disable_cache = disable_cache

    def _download(
        self,
        file_url: str,
        filename: str,
        response_ok: t.Callable[[requests.Response], bool],
    ) -> str:
        assert self._cache_dir is not None
        os.makedirs(self._cache_dir, exist_ok=True)
        dest = os.path.join(self._cache_dir, filename)

        def check_response_for_download(r: requests.Response) -> bool:
            # if the response indicates a cache hit, treat it as valid
            # this ensures that we short-circuit any further evaluation immediately on
            # a hit
            if _cache_hit(dest, r):
                return True
            # we now know it's not a hit, so validate the content (forces download)
            return response_ok(r)

        response = _get_request(file_url, response_ok=check_response_for_download)
        # check to see if we have a file which matches the connection
        # only download if we do not (cache miss, vs hit)
        if not _cache_hit(dest, response):
            _atomic_write(dest, response.content)

        return dest

    @contextlib.contextmanager
    def open(
        self,
        file_url: str,
        filename: str,
        validate_response: t.Callable[[requests.Response], bool],
    ) -> t.Iterator[t.IO[bytes]]:
        if (not self._cache_dir) or self._disable_cache:
            yield io.BytesIO(
                _get_request(file_url, response_ok=validate_response).content
            )
        else:
            with open(
                self._download(file_url, filename, response_ok=validate_response), "rb"
            ) as fp:
                yield fp

    def bind(
        self,
        file_url: str,
        filename: str | None = None,
        validation_callback: t.Callable[[bytes], t.Any] | None = None,
    ) -> BoundCacheDownloader:
        return BoundCacheDownloader(
            file_url, self, filename=filename, validation_callback=validation_callback
        )


class BoundCacheDownloader:
    def __init__(
        self,
        file_url: str,
        downloader: CacheDownloader,
        *,
        filename: str | None = None,
        validation_callback: t.Callable[[bytes], t.Any] | None = None,
    ) -> None:
        self._file_url = file_url
        self._filename = filename or url_to_cache_filename(file_url)
        self._downloader = downloader
        self._validation_callback = validation_callback

    @contextlib.contextmanager
    def open(self) -> t.Iterator[t.IO[bytes]]:
        with self._downloader.open(
            self._file_url,
            self._filename,
            validate_response=self._validate_response,
        ) as fp:
            yield fp

    def _validate_response(self, response: requests.Response) -> bool:
        if not self._validation_callback:
            return True

        try:
            self._validation_callback(response.content)
            return True
        except ValueError:
            return False



import json

import pytest

TITLE_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {"title": {"type": "string"}},
    "required": ["title"],
}


def test_non_json_instance(run_line, tmp_path):
    schema = tmp_path / "schema.json"
    instance = tmp_path / "instance.json"
    schema.write_text("{}")
    instance.write_text("{")

    res = run_line(["check-jsonschema", "--schemafile", str(schema), str(instance)])
    assert res.exit_code == 1
    assert f"Failed to parse {str(instance)}" in res.stdout


@pytest.mark.parametrize("outformat", ["TEXT", "JSON"])
def test_non_json_instance_mixed_with_valid_data(run_line, tmp_path, outformat):
    schema = tmp_path / "schema.json"
    malformed_instance = tmp_path / "malformed_instance.json"
    good_instance = tmp_path / "good_instance.json"
    schema.write_text(json.dumps(TITLE_SCHEMA))
    malformed_instance.write_text("{")
    good_instance.write_text('{"title": "ohai"}')

    res = run_line(
        [
            "check-jsonschema",
            "-o",
            outformat,
            "--schemafile",
            str(schema),
            str(malformed_instance),
            str(good_instance),
        ]
    )
    assert res.exit_code == 1
    if outformat == "TEXT":
        assert f"Failed to parse {str(malformed_instance)}" in res.stdout
    else:
        report = json.loads(res.stdout)
        assert report["status"] == "fail"
        assert "errors" in report
        assert report["errors"] == []
        assert "parse_errors" in report
        assert len(report["parse_errors"]) == 1
        error_item = report["parse_errors"][0]
        assert error_item["filename"] == str(malformed_instance)
        assert f"Failed to parse {str(malformed_instance)}" in error_item["message"]


@pytest.mark.parametrize("outformat", ["TEXT", "JSON"])
def test_non_json_instance_mixed_with_valid_and_invalid_data(
    run_line, tmp_path, outformat
):
    schema = tmp_path / "schema.json"
    malformed_instance = tmp_path / "malformed_instance.json"
    good_instance = tmp_path / "good_instance.json"
    bad_instance = tmp_path / "bad_instance.json"
    schema.write_text(json.dumps(TITLE_SCHEMA))
    malformed_instance.write_text("{")
    good_instance.write_text('{"title": "ohai"}')
    bad_instance.write_text('{"title": false}')

    res = run_line(
        [
            "check-jsonschema",
            "-o",
            outformat,
            "--schemafile",
            str(schema),
            str(good_instance),
            str(malformed_instance),
            str(bad_instance),
        ]
    )
    assert res.exit_code == 1
    if outformat == "TEXT":
        assert f"Failed to parse {str(malformed_instance)}" in res.stdout
        assert (
            f"{str(bad_instance)}::$.title: False is not of type 'string'" in res.stdout
        )
    else:
        report = json.loads(res.stdout)
        assert report["status"] == "fail"

        assert "errors" in report
        assert len(report["errors"]) == 1

        assert "parse_errors" in report
        assert len(report["parse_errors"]) == 1
        error_item = report["parse_errors"][0]
        assert error_item["filename"] == str(malformed_instance)
        assert f"Failed to parse {str(malformed_instance)}" in error_item["message"]



import json

import pytest
import responses

CASES = {
    "case1": {
        "main_schema": {
            "$schema": "http://json-schema.org/draft-07/schema",
            "properties": {
                "title": {"$ref": "./title_schema.json"},
            },
            "additionalProperties": False,
        },
        "other_schemas": {"title_schema": {"type": "string"}},
        "passing_document": {"title": "doc one"},
        "failing_document": {"title": 2},
    },
    "case2": {
        "main_schema": {
            "$schema": "http://json-schema.org/draft-07/schema",
            "type": "object",
            "required": ["test"],
            "properties": {"test": {"$ref": "./values.json#/$defs/test"}},
        },
        "other_schemas": {
            "values": {
                "$schema": "http://json-schema.org/draft-07/schema",
                "$defs": {"test": {"type": "string"}},
            }
        },
        "passing_document": {"test": "some data"},
        "failing_document": {"test": {"foo": "bar"}},
    },
}


@pytest.mark.parametrize("check_passes", (True, False))
@pytest.mark.parametrize("casename", ("case1", "case2"))
def test_remote_ref_resolution_simple_case(run_line, check_passes, casename, tmp_path):
    main_schema_loc = "https://example.com/main.json"
    responses.add("GET", main_schema_loc, json=CASES[casename]["main_schema"])
    for name, subschema in CASES[casename]["other_schemas"].items():
        other_schema_loc = f"https://example.com/{name}.json"
        responses.add("GET", other_schema_loc, json=subschema)

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(
        json.dumps(
            CASES[casename]["passing_document"]
            if check_passes
            else CASES[casename]["failing_document"]
        )
    )

    result = run_line(
        ["check-jsonschema", "--schemafile", main_schema_loc, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


@pytest.mark.parametrize("casename", ("case1", "case2"))
@pytest.mark.parametrize("disable_cache", (True, False))
def test_remote_ref_resolution_cache_control(
    run_line, tmp_path, get_ref_cache_loc, casename, disable_cache
):
    main_schema_loc = "https://example.com/main.json"
    responses.add("GET", main_schema_loc, json=CASES[casename]["main_schema"])

    ref_locs = []
    for name, subschema in CASES[casename]["other_schemas"].items():
        other_schema_loc = f"https://example.com/{name}.json"
        responses.add("GET", other_schema_loc, json=subschema)
        ref_locs.append(other_schema_loc)

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps(CASES[casename]["passing_document"]))

    # run the command
    result = run_line(
        ["check-jsonschema", "--schemafile", main_schema_loc, str(instance_path)]
        + (["--no-cache"] if disable_cache else [])
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    assert result.exit_code == 0, output

    cache_locs = []
    for ref_loc in ref_locs:
        cache_locs.append(get_ref_cache_loc(ref_loc))
    assert cache_locs  # sanity check
    if disable_cache:
        for loc in cache_locs:
            assert not loc.exists()
    else:
        for loc in cache_locs:
            assert loc.exists()


@pytest.mark.parametrize("casename", ("case1", "case2"))
@pytest.mark.parametrize("check_passes", (True, False))
def test_remote_ref_resolution_loads_from_cache(
    run_line, tmp_path, get_ref_cache_loc, inject_cached_ref, casename, check_passes
):
    main_schema_loc = "https://example.com/main.json"
    responses.add("GET", main_schema_loc, json=CASES[casename]["main_schema"])

    ref_locs = []
    cache_locs = []
    for name, subschema in CASES[casename]["other_schemas"].items():
        other_schema_loc = f"https://example.com/{name}.json"
        # intentionally populate the HTTP location with "bad data"
        responses.add("GET", other_schema_loc, json="{}")
        ref_locs.append(other_schema_loc)

        # but populate the cache with "good data"
        inject_cached_ref(other_schema_loc, json.dumps(subschema))
        cache_locs.append(get_ref_cache_loc(other_schema_loc))

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(
        json.dumps(
            CASES[casename]["passing_document"]
            if check_passes
            else CASES[casename]["failing_document"]
        )
    )

    # run the command
    result = run_line(
        ["check-jsonschema", "--schemafile", main_schema_loc, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


# this test ensures that `$id` is preferred for the base URI over
# the retrieval URI
@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_prefers_id_over_retrieval_uri(run_line, tmp_path, check_passes):
    main_schema = {
        "$id": "https://example.org/schemas/main.json",
        "$schema": "http://json-schema.org/draft-07/schema",
        "properties": {
            "title": {"$ref": "./title_schema.json"},
        },
        "additionalProperties": False,
    }
    title_schema = {"type": "string"}

    retrieval_uri = "https://example.org/alternate-path-retrieval-only/schemas/main"
    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add(
        "GET", "https://example.org/schemas/title_schema.json", json=title_schema
    )

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        ["check-jsonschema", "--schemafile", retrieval_uri, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_does_not_callout_for_absolute_ref_to_retrieval_uri(
    run_line, tmp_path, check_passes
):
    retrieval_uri = "https://example.net/schemas/main"

    main_schema = {
        "$id": "https://example.net/schemas/some-uri-which-will-never-be-used/main.json",
        "$schema": "http://json-schema.org/draft-07/schema",
        "$defs": {"title": {"type": "string"}},
        "properties": {
            "title": {"$ref": f"{retrieval_uri}#/$defs/title"},
        },
        "additionalProperties": False,
    }

    # exactly one GET to the retrieval URI will work
    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add("GET", retrieval_uri, json={"error": "permafrost melted"}, status=500)

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        ["check-jsonschema", "--schemafile", retrieval_uri, str(instance_path)]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


# this test ensures that `$id` is overwritten when `--base-uri` is used
@pytest.mark.parametrize("check_passes", (True, False))
def test_ref_resolution_with_custom_base_uri(run_line, tmp_path, check_passes):
    retrieval_uri = "https://example.org/retrieval-and-in-schema-only/schemas/main"
    explicit_base_uri = "https://example.org/schemas/main"
    main_schema = {
        "$id": retrieval_uri,
        "$schema": "http://json-schema.org/draft-07/schema",
        "properties": {
            "title": {"$ref": "./title_schema.json"},
        },
        "additionalProperties": False,
    }
    title_schema = {"type": "string"}

    responses.add("GET", retrieval_uri, json=main_schema)
    responses.add(
        "GET", "https://example.org/schemas/title_schema.json", json=title_schema
    )

    instance_path = tmp_path / "instance.json"
    instance_path.write_text(json.dumps({"title": "doc one" if check_passes else 2}))

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            retrieval_uri,
            "--base-uri",
            explicit_base_uri,
            str(instance_path),
        ]
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output


@pytest.mark.parametrize("num_instances", (1, 2, 10))
@pytest.mark.parametrize("check_passes", (True, False))
def test_remote_ref_resolution_callout_count_is_scale_free_in_instancefiles(
    run_line, tmp_path, num_instances, check_passes
):
    """
    Test that for any N > 1, validation of a schema with a ref against N instance files
    has exactly the same number of callouts as validation when N=1

    This proves that the validator and caching are working correctly, and we aren't
    repeating callouts to rebuild state.
    """
    schema_uri = "https://example.org/schemas/main.json"
    ref_uri = "https://example.org/schemas/title_schema.json"

    main_schema = {
        "$id": schema_uri,
        "$schema": "http://json-schema.org/draft-07/schema",
        "properties": {
            "title": {"$ref": "./title_schema.json"},
        },
        "additionalProperties": False,
    }
    title_schema = {"type": "string"}
    responses.add("GET", schema_uri, json=main_schema)
    responses.add("GET", ref_uri, json=title_schema)

    # write N documents
    instance_doc = {"title": "doc one" if check_passes else 2}
    instance_paths = []
    for i in range(num_instances):
        instance_path = tmp_path / f"instance{i}.json"
        instance_path.write_text(json.dumps(instance_doc))
        instance_paths.append(str(instance_path))

    result = run_line(
        [
            "check-jsonschema",
            "--schemafile",
            schema_uri,
        ]
        + instance_paths
    )
    output = f"\nstdout:\n{result.stdout}\n\nstderr:\n{result.stderr}"
    if check_passes:
        assert result.exit_code == 0, output
    else:
        assert result.exit_code == 1, output

    # this is the moment of the "real" test run here:
    # no matter how many instances there were, there should only have been two calls
    # (one for the schema and one for the $ref)
    assert len(responses.calls) == 2
    assert len([c for c in responses.calls if c.request.url == schema_uri]) == 1
    assert len([c for c in responses.calls if c.request.url == ref_uri]) == 1



import pytest


def test_checker_non_json_schemafile(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text("{")
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", str(foo), str(bar)])
    assert res.exit_code == 1
    assert "schemafile could not be parsed" in res.stderr


def test_checker_invalid_schemafile(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text('{"title": {"foo": "bar"}}')
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", str(foo), str(bar)])
    assert res.exit_code == 1
    assert "schemafile was not valid" in res.stderr


def test_checker_invalid_schemafile_scheme(run_line, tmp_path):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    foo.write_text('{"title": "foo"}')
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", f"ftp://{foo}", str(bar)])
    assert res.exit_code == 1
    assert "only supports http, https" in res.stderr


@pytest.mark.parametrize(
    "add_args",
    [
        pytest.param([], id="noargs"),
        # ensure that this works even when regex checking is disabled
        pytest.param(["--disable-formats", "*"], id="all-formats-disabled"),
        pytest.param(["--disable-formats", "regex"], id="regex-format-disabled"),
    ],
)
def test_checker_invalid_schemafile_due_to_bad_regex(run_line, tmp_path, add_args):
    foo = tmp_path / "foo.json"
    bar = tmp_path / "bar.json"
    # too many backslash escapes -- not a valid Unicode-mode regex
    foo.write_text(r'{"properties": {"foo": {"pattern": "\\\\p{N}"}}}')
    bar.write_text("{}")

    res = run_line(["check-jsonschema", "--schemafile", str(foo), str(bar), *add_args])
    assert res.exit_code == 1
    assert "schemafile was not valid" in res.stderr



def test_gitlab_reference_handling_on_bad_data(run_line, tmp_path):
    doc = tmp_path / "data.yml"
    doc.write_text("""\
include:
  - local: setup.yml

test:
  script:
    # !reference not a list, error
    - !reference .setup
    - echo running my own command
""")

    res = run_line(
        [
            "check-jsonschema",
            "--builtin-schema",
            "gitlab-ci",
            "--data-transform",
            "gitlab-ci",
            str(doc),
        ],
        catch_exceptions=True,
    )
    assert res.exit_code == 1



import json

FORMAT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {
        "title": {
            "type": "string",
        },
        "date": {
            "type": "string",
            "format": "date",
        },
    },
}

PASSING_DOCUMENT = {
    "title": "doc one",
    "date": "2021-10-28",
}

FAILING_DOCUMENT = {
    "title": "doc one",
    "date": "foo",
}


def test_format_check_passing(run_line_simple, tmp_path):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc1 = tmp_path / "doc1.json"
    doc1.write_text(json.dumps(PASSING_DOCUMENT))

    run_line_simple(["--schemafile", str(schemafile), str(doc1)])


def test_format_failure_exit_error(run_line, tmp_path):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc1 = tmp_path / "doc1.json"
    doc1.write_text(json.dumps(FAILING_DOCUMENT))

    res = run_line(["check-jsonschema", "--schemafile", str(schemafile), str(doc1)])
    assert res.exit_code == 1


def test_format_failure_ignore(run_line_simple, tmp_path):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc1 = tmp_path / "doc1.json"
    doc1.write_text(json.dumps(FAILING_DOCUMENT))

    run_line_simple(
        [
            "--disable-formats",
            "*",
            "--schemafile",
            str(schemafile),
            str(doc1),
        ]
    )


def test_format_failure_ignore_multidoc(run_line_simple, tmp_path):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc1 = tmp_path / "doc1.json"
    doc1.write_text(json.dumps(FAILING_DOCUMENT))

    doc2 = tmp_path / "doc2.json"
    doc2.write_text(json.dumps(PASSING_DOCUMENT))

    run_line_simple(
        [
            "--disable-formats",
            "*",
            "--schemafile",
            str(schemafile),
            str(doc1),
            str(doc2),
        ]
    )



from __future__ import annotations

from unittest import mock

import click
import pytest

from check_jsonschema import main as cli_main
from check_jsonschema.cli.parse_result import ParseResult, SchemaLoadingMode


class BoxedContext:
    ref = None


def touch_files(dirpath, *filenames):
    for fname in filenames:
        (dirpath / fname).touch()


@pytest.fixture
def boxed_context():
    return BoxedContext()


@pytest.fixture
def mock_parse_result():
    args = ParseResult()
    with mock.patch("check_jsonschema.cli.main_command.ParseResult") as m:
        m.return_value = args
        yield args


@pytest.fixture(autouse=True)
def mock_cli_exec(boxed_context):
    def get_ctx(*args):
        boxed_context.ref = click.get_current_context()

    with mock.patch(
        "check_jsonschema.cli.main_command.execute", side_effect=get_ctx
    ) as m:
        yield m


@pytest.mark.parametrize(
    "schemafile,builtin_schema,check_metaschema,expect_mode",
    [
        ("foo.json", None, False, SchemaLoadingMode.filepath),
        (None, "foo", False, SchemaLoadingMode.builtin),
        (None, None, True, SchemaLoadingMode.metaschema),
    ],
)
def test_parse_result_set_schema(
    schemafile, builtin_schema, check_metaschema, expect_mode
):
    args = ParseResult()
    # starts as None (always)
    assert args.schema_path is None

    args.set_schema(schemafile, builtin_schema, check_metaschema)
    assert args.schema_mode == expect_mode
    if schemafile:
        assert args.schema_path == schemafile
    if builtin_schema:
        assert args.schema_path == builtin_schema
    if check_metaschema:
        assert args.schema_path is None


def test_requires_some_args(cli_runner):
    result = cli_runner.invoke(cli_main, [])
    assert result.exit_code == 2


def test_schemafile_and_instancefile(
    cli_runner, mock_parse_result, in_tmp_dir, tmp_path
):
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(cli_main, ["--schemafile", "schema.json", "foo.json"])
    assert mock_parse_result.schema_mode == SchemaLoadingMode.filepath
    assert mock_parse_result.schema_path == "schema.json"
    assert isinstance(mock_parse_result.instancefiles, tuple)
    for f in mock_parse_result.instancefiles:
        assert isinstance(f, click.utils.LazyFile)
    assert tuple(f.name for f in mock_parse_result.instancefiles) == ("foo.json",)


def test_requires_at_least_one_instancefile(cli_runner):
    result = cli_runner.invoke(cli_main, ["--schemafile", "schema.json"])
    assert result.exit_code == 2


def test_requires_schemafile(cli_runner, in_tmp_dir, tmp_path):
    touch_files(tmp_path, "foo.json")
    result = cli_runner.invoke(cli_main, ["foo.json"])
    assert result.exit_code == 2


def test_no_cache_defaults_false(cli_runner, mock_parse_result):
    cli_runner.invoke(cli_main, ["--schemafile", "schema.json", "foo.json"])
    assert mock_parse_result.disable_cache is False


def test_no_cache_flag_is_true(cli_runner, mock_parse_result, in_tmp_dir, tmp_path):
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(
        cli_main, ["--schemafile", "schema.json", "foo.json", "--no-cache"]
    )
    assert mock_parse_result.disable_cache is True


@pytest.mark.parametrize(
    "cmd_args",
    [
        [
            "--schemafile",
            "x.json",
            "--builtin-schema",
            "vendor.travis",
        ],
        [
            "--schemafile",
            "x.json",
            "--builtin-schema",
            "vendor.travis",
            "--check-metaschema",
        ],
        [
            "--schemafile",
            "x.json",
            "--check-metaschema",
        ],
        [
            "--builtin-schema",
            "vendor.travis",
            "--check-metaschema",
        ],
    ],
)
def test_mutex_schema_opts(cli_runner, cmd_args, in_tmp_dir, tmp_path):
    touch_files(tmp_path, "foo.json")
    result = cli_runner.invoke(cli_main, cmd_args + ["foo.json"])
    assert result.exit_code == 2
    assert "are mutually exclusive" in result.stderr


@pytest.mark.parametrize(
    "cmd_args",
    [
        ["--version"],
        ["--help"],
        ["-h"],
    ],
)
def test_supports_common_option(cli_runner, cmd_args):
    result = cli_runner.invoke(cli_main, cmd_args)
    assert result.exit_code == 0


@pytest.mark.parametrize(
    "setting,expect_value", [(None, None), ("1", False), ("0", False)]
)
def test_no_color_env_var(
    cli_runner, monkeypatch, setting, expect_value, boxed_context, in_tmp_dir, tmp_path
):
    if setting is None:
        monkeypatch.delenv("NO_COLOR", raising=False)
    else:
        monkeypatch.setenv("NO_COLOR", setting)

    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(cli_main, ["--schemafile", "schema.json", "foo.json"])
    assert boxed_context.ref.color == expect_value


@pytest.mark.parametrize(
    "setting,expected_value",
    [(None, None), ("auto", None), ("always", True), ("never", False)],
)
def test_color_cli_option(
    cli_runner, setting, expected_value, boxed_context, in_tmp_dir, tmp_path
):
    args = ["--schemafile", "schema.json", "foo.json"]
    if setting:
        args.extend(("--color", setting))
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(cli_main, args)
    assert boxed_context.ref.color == expected_value


def test_no_color_env_var_overrides_cli_option(
    cli_runner, monkeypatch, mock_cli_exec, boxed_context, in_tmp_dir, tmp_path
):
    monkeypatch.setenv("NO_COLOR", "1")
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(
        cli_main, ["--color=always", "--schemafile", "schema.json", "foo.json"]
    )
    assert boxed_context.ref.color is False


@pytest.mark.parametrize(
    "setting,expected_value",
    [("auto", 0), ("always", 0), ("never", 0), ("anything_else", 2)],
)
def test_color_cli_option_is_choice(
    cli_runner, setting, expected_value, in_tmp_dir, tmp_path
):
    touch_files(tmp_path, "foo.json")
    assert (
        cli_runner.invoke(
            cli_main,
            ["--color", setting, "--schemafile", "schema.json", "foo.json"],
        ).exit_code
        == expected_value
    )


def test_formats_default_to_enabled(
    cli_runner, mock_parse_result, in_tmp_dir, tmp_path
):
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(cli_main, ["--schemafile", "schema.json", "foo.json"])
    assert mock_parse_result.disable_all_formats is False
    assert mock_parse_result.disable_formats == ()


@pytest.mark.parametrize(
    "addargs",
    (
        [
            "--disable-formats",
            "uri-reference",
            "--disable-formats",
            "date-time",
        ],
        ["--disable-formats", "uri-reference,date-time"],
    ),
)
def test_disable_selected_formats(
    cli_runner, mock_parse_result, addargs, in_tmp_dir, tmp_path
):
    touch_files(tmp_path, "foo.json")
    cli_runner.invoke(
        cli_main,
        [
            "--schemafile",
            "schema.json",
            "foo.json",
        ]
        + addargs,
    )
    assert mock_parse_result.disable_all_formats is False
    assert set(mock_parse_result.disable_formats) == {"uri-reference", "date-time"}


@pytest.mark.parametrize(
    "addargs",
    (
        [
            "--disable-formats",
            "uri-reference",
            "--disable-formats",
            "date-time",
            "--disable-formats",
            "*",
        ],
        ["--disable-formats", "*"],
        ["--disable-formats", "*,email"],
    ),
)
def test_disable_all_formats(
    cli_runner, mock_parse_result, addargs, in_tmp_dir, tmp_path
):
    touch_files(tmp_path, "foo.json")
    # this should be an override, with or without other args
    cli_runner.invoke(
        cli_main,
        [
            "--schemafile",
            "schema.json",
            "foo.json",
        ]
        + addargs,
    )
    assert mock_parse_result.disable_all_formats is True


def test_can_specify_custom_validator_class(
    cli_runner, mock_parse_result, mock_module, in_tmp_dir, tmp_path
):
    mock_module("foo.py", "class MyValidator: pass")
    import foo

    touch_files(tmp_path, "foo.json")
    result = cli_runner.invoke(
        cli_main,
        [
            "--schemafile",
            "schema.json",
            "foo.json",
            "--validator-class",
            "foo:MyValidator",
        ],
    )
    assert result.exit_code == 0
    assert mock_parse_result.validator_class == foo.MyValidator


@pytest.mark.parametrize(
    "failmode", ("syntax", "import", "attr", "function", "non_callable")
)
def test_custom_validator_class_fails(
    cli_runner, mock_parse_result, mock_module, failmode, in_tmp_dir, tmp_path
):
    mock_module(
        "foo.py",
        """\
class MyValidator: pass

def validator_func(*args, **kwargs):
    return MyValidator(*args, **kwargs)

other_thing = 100
""",
    )

    if failmode == "syntax":
        arg = "foo.MyValidator"
    elif failmode == "import":
        arg = "foo.bar:MyValidator"
    elif failmode == "attr":
        arg = "foo:no_such_attr"
    elif failmode == "function":
        arg = "foo:validator_func"
    elif failmode == "non_callable":
        arg = "foo:other_thing"
    else:
        raise NotImplementedError

    touch_files(tmp_path, "foo.json")
    result = cli_runner.invoke(
        cli_main,
        ["--schemafile", "schema.json", "foo.json", "--validator-class", arg],
    )
    assert result.exit_code == 2

    if failmode == "syntax":
        assert "is not a valid specifier" in result.stderr
    elif failmode == "import":
        assert "was not an importable module" in result.stderr
    elif failmode == "attr":
        assert "was not resolvable to a class" in result.stderr
    elif failmode in ("function", "non_callable"):
        assert "is not a class" in result.stderr
    else:
        raise NotImplementedError
